/* (C) Copyright 1998 Florian La Roche
 * for details see file COPYING
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <net/if.h>
#include <net/if_arp.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include "bootpc.h"

/*
 * A bootp server might send back a larger packet, if the original size
 * is not enough. We can receive packets that are this much larger.
 */
#define BOOTP_ADD 1024

struct bootp {
	unsigned char	bp_op;		/* message type */
	unsigned char	bp_htype;	/* hardware addr type */
	unsigned char	bp_hlen;	/* hardware addr length */
	unsigned char	bp_hops;	/* gateway hops */
	__u32		bp_xid;		/* transaction ID */
	unsigned short	bp_secs;	/* seconds since boot began */
	unsigned short	bp_flags;	/* RFC1542 flags field */
	struct in_addr	bp_ciaddr;	/* client IP address */
	struct in_addr	bp_yiaddr;	/* 'your' IP address */
	struct in_addr	bp_siaddr;	/* server IP address */
	struct in_addr	bp_giaddr;	/* gateway IP address */
	unsigned char	bp_chaddr[16];	/* client hardware address */
	unsigned char	bp_sname[64];	/* server host name */
	unsigned char	bp_file[128];	/* boot file name */
	unsigned char	bp_vend[64];	/* vendor-specific area */

	/* bootp server may send a larger packet back */
	unsigned char	dummy[BOOTP_ADD];
};

/* four byte cookie and end tag */
static unsigned char mincookie[] = { 99, 130, 83, 99, 255 };

int do_bootp (const char *device, int flags, __u32 *ip)
{
	int sfd, i, retry_wait, waited = 0, received = 0, rancopy = 0;
	struct sockaddr_in cli_addr, serv_addr;
	struct bootp pkg;
	struct ifreq ifr;
	struct timeval tp;
	struct timeval timeout;

	/* set the destination address */
	memset (&serv_addr, 0, sizeof (serv_addr));
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_addr.s_addr = bootp_server;
	serv_addr.sin_port = htons (67);	/* upd port for bootp server */

	/* set the source address */
	memset (&cli_addr, 0, sizeof (cli_addr));
	cli_addr.sin_family = AF_INET;
	cli_addr.sin_addr.s_addr = htonl (INADDR_ANY);
	cli_addr.sin_port = htons (68);		/* udp port for bootp client */

	if ((sfd = socket (AF_INET, SOCK_DGRAM, 0)) == -1)
		goto error;
	i = 1;
	if (setsockopt (sfd, SOL_SOCKET, SO_BROADCAST, &i, sizeof (i)) < 0)
		goto error;
	if (bind (sfd, (struct sockaddr *)&cli_addr, sizeof(cli_addr)) < 0)
		goto error;
	/* XXX move the detection of the hw address out of this function? */
	if ((i = strlen(device) + 1) > sizeof(ifr.ifr_name))
		goto error;
	memcpy (ifr.ifr_name, device, i);
	if (ioctl (sfd, SIOCGIFHWADDR, &ifr) < 0)
		goto error;
	memset (&pkg, 0, sizeof (pkg));
	pkg.bp_htype = ifr.ifr_hwaddr.sa_family;
	/* XXX can this be done better ? */
	pkg.bp_hlen = IFHWADDRLEN;
	pkg.bp_op = 1;	/* this is a bootp request, 2 is a reply */
	pkg.bp_flags = flags;

	/* XXX we could use /dev/random */
	srand ((gettimeofday (&tp, NULL) == -1) ? 0 : tp.tv_sec + tp.tv_usec);

	switch (pkg.bp_htype) {
	case ARPHRD_ARCNET:
		memcpy (pkg.bp_chaddr + IFHWADDRLEN - 1,
				ifr.ifr_hwaddr.sa_data, 1);
		pkg.bp_htype = ARPHRD_ETHER;
		pkg.bp_hlen = IFHWADDRLEN;
		break;
	default:
		memcpy (pkg.bp_chaddr, ifr.ifr_hwaddr.sa_data, IFHWADDRLEN);
	}
	memcpy (pkg.bp_vend, mincookie, sizeof (mincookie));

	retry_wait = 2;
	while (retry_wait <= bootp_timeout) {
		fd_set rfds;

		if (!received) {
			pkg.bp_xid = rancopy = rand ();
			if (sendto (sfd, &pkg, sizeof(pkg) - BOOTP_ADD, 0,
				(struct sockaddr *) &serv_addr,
				sizeof (serv_addr)) < 0)
				goto error;
			timeout.tv_sec = retry_wait +
					(1 + (rand() & (retry_wait - 1)));
			waited += timeout.tv_sec;
			timeout.tv_usec = 0;
		}
		FD_ZERO (&rfds);
		FD_SET (sfd, &rfds);
		i = select (sfd + 1, &rfds, NULL, NULL, &timeout);
		if (i < 0) goto error;
		if (!FD_ISSET (sfd, &rfds)) {
			pkg.bp_secs = htons(waited);
			retry_wait = retry_wait * 2;
			received = 0;
			continue;
		}
		i = recvfrom (sfd, &pkg, sizeof (pkg), 0, NULL, NULL);
		if (i < 0) goto error;
		received = 1;
		if (i < (sizeof(pkg)-BOOTP_ADD) || pkg.bp_xid != rancopy)
			continue;		/* ignore packet */
		if (memcmp(pkg.bp_vend, mincookie, sizeof(mincookie) - 1))
			continue;		/* ignore packet */
		break;
	}
	close (sfd);
	if (!received)
		return -1;
	*ip = pkg.bp_yiaddr.s_addr;
	parse_options (pkg.bp_vend,
			i + sizeof(pkg.bp_vend) + BOOTP_ADD - sizeof(pkg), *ip);
	return 0;

error:
	if (sfd != -1)
		close(sfd);
	return -1;
}
